import logging
import time
from concurrent.futures import ThreadPoolExecutor

import grpc
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai_utils.auth.helper import ClarifaiAuthHelper
from clarifai_utils.client.abc import RpcCallable, V2Stub

throttle_status_codes = {
    status_code_pb2.CONN_THROTTLED,
    status_code_pb2.CONN_EXCEED_HOURLY_LIMIT,
}

retry_codes_grpc = {
    grpc.StatusCode.UNAVAILABLE,
}

_threadpool = ThreadPoolExecutor(100)


def create_stub(auth_helper=None, max_retry_attempts=10):
  '''
  Create client stub that handles authorization and basic retries for
  unavailable or throttled connections.

  Args:
    auth_helper:  ClarifaiAuthHelper to use for auth metadata (default: from env)
    max_retry_attempts:  max attempts to retry rpcs with retryable failures
  '''
  stub = AuthorizedStub(auth_helper)
  if max_retry_attempts > 0:
    return RetryStub(stub, max_retry_attempts)
  return stub


class AuthorizedStub(V2Stub):
  '''
  V2Stub proxy that inserts metadata authorization in rpc calls.
  '''

  def __init__(self, auth_helper: ClarifaiAuthHelper = None):
    if auth_helper is None:
      auth_helper = ClarifaiAuthHelper.from_env()
    self.stub = auth_helper.get_stub()
    self.metadata = auth_helper.metadata

  def __getattr__(self, name):
    value = getattr(self.stub, name)
    if isinstance(value, RpcCallable):
      value = _AuthorizedRpcCallable(value, self.metadata)
    return value


class _AuthorizedRpcCallable(RpcCallable):

  def __init__(self, func, metadata):
    self.f = func
    self.metadata = metadata

  def __repr__(self):
    return repr(self.f)

  def __call__(self, *args, **kwargs):
    metadata = kwargs.pop('metadata', self.metadata)
    return self.f(*args, **kwargs, metadata=metadata)

  def future(self, *args, **kwargs):
    metadata = kwargs.pop('metadata', self.metadata)
    return self.f.future(*args, **kwargs, metadata=metadata)

  def __getattr__(self, name):
    return getattr(self.f, name)


class RetryStub(V2Stub):
  '''
  V2Stub proxy that retries requests (currently on unavailable server or throttle codes)
  '''

  def __init__(self, stub, max_attempts=10, backoff_time=5):
    self.stub = stub
    self.max_attempts = max_attempts
    self.backoff_time = backoff_time

  def __getattr__(self, name):
    value = getattr(self.stub, name)
    if isinstance(value, RpcCallable):
      value = _RetryRpcCallable(value, self.max_attempts, self.backoff_time)
    return value


class _RetryRpcCallable(RpcCallable):

  def __init__(self, func, max_attempts, backoff_time):
    self.f = func
    self.max_attempts = max_attempts
    self.backoff_time = backoff_time

  def __repr__(self):
    return repr(self.f)

  def __call__(self, *args, **kwargs):
    attempt = 0
    while attempt < self.max_attempts:
      attempt += 1
      if attempt != 1:
        time.sleep(self.backoff_time)  # TODO better backoff between attempts
      try:
        response = self.f(*args, **kwargs)
        if (response.status.code in throttle_status_codes) and attempt < self.max_attempts:
          logging.debug('Retrying with status %s' % str(response.status))
        else:
          return response
      except grpc.RpcError as e:
        if (e.code() in retry_codes_grpc) and attempt < self.max_attempts:
          logging.debug('Retrying with status %s' % e.code())
        else:
          raise

  def future(self, *args, **kwargs):
    # TODO use single result event loop thread with asyncio
    return _threadpool.submit(self, *args, **kwargs)

  def __getattr__(self, name):
    return getattr(self.f, name)



from typing import Dict, List


class ClarifaiDataset:
  """
  Dataset base class
  """

  def __init__(self, csv_file_path: str, dataset_id: str, split: str) -> None:
    self.csv_file_path = csv_file_path
    self.dataset_id = dataset_id
    self.split = split
    self._all_input_protos = []

  def __len__(self):
    """
    Get size of all input protos
    """
    return len(self._all_input_protos)

  def to_list(self, input_protos: List):
    """
    Return a list of input protos.
    """
    return list(input_protos)

  def create_input_protos(self,
                          image_path: str,
                          label: str,
                          input_id: str,
                          dataset_id: str,
                          metadata: Dict,
                          use_urls=False,
                          allow_dups=False):
    """
    Create input protos for each image, label input pair.
    Args:
    	`image_path`: full image path or valid url with
    		ending with an image extension.
    	`label`: image label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: image metadata
    	`use_urls`: If set to True it means all image_paths are provided as urls and
    		hence uploading will attempt to read images from urls.
    		The default behavior is reading images from local storage.
    	`allow_dups`: Boolean indicating whether to allow duplicate url inputs
    Returns:
    	An input proto representing a single row input
    """
    raise NotImplementedError()

  def _get_input_protos(self) -> List:
    """
    Create input protos for each row of the dataframe.
    Returns:
    	A list of input protos
    """
    raise NotImplementedError()


class Chunker:
  """
  Split an input sequence into small chunks.
  """

  def __init__(self, seq: List, size: int) -> None:
    self.seq = seq
    self.size = size

  def chunk(self):
    """
    Chunk input sequence.
    """
    return [self.seq[pos:pos + self.size] for pos in range(0, len(self.seq), self.size)]



import gc
import os
from typing import List, Tuple
from xml.etree import ElementTree as et
import numpy as np
import pandas as pd


def validate_bboxes(bboxes: List) -> List:
  """
  Validate bounding boxes for correctness of dimensions.
  Args:
  	`bboxes`: a list holding a single bounding box in the format,
		 [x_min, y_min, x_max, y_max]
  """
  # Some bbox annotations may have their min dims > max dims
  # Check for these as clarifai annotation post requests will throw errors
  x_min, y_min, x_max, y_max = bboxes

  #assert max_dims > min_dims
  assert x_max > x_min, f"x_max must be greater than x_min. Got x_max:{x_max} x_min:{x_min}"
  assert y_max > y_min, f"y_max must be greater than y_min. Got y_max:{y_max} y_min:{y_min}"

  top_row = [x_min, y_min]
  bottom_row = [x_max, y_max]

  # bound bbox coords between 0 and 1
  top_row = list(np.clip(top_row, 0., 1.))
  bottom_row = list(np.clip(bottom_row, 0., 1.))

  assert bottom_row[0] > top_row[0], f"x_max must be greater than x_min. Got x_max:{bottom_row[0]} x_min:{top_row[0]}"
  assert bottom_row[1] > top_row[1], f"y_max must be greater than y_min. Got y_max:{bottom_row[1]} y_min:{top_row[1]}"

  return top_row + bottom_row


def load_xml_annotations(xml_file: str) -> Tuple:
  """
  Read image, labels and annotations.
  """
  tree = et.parse(xml_file)
  root = tree.getroot()

  all_bboxes = []
  labels = []

  for boxes in root.iter('object'):
    labels.append(boxes.find('name').text)
    img_name = root.find('filename').text
    # image dims relative to the annotator
    image_width = int(root.find('size').find('width').text)
    image_height = int(root.find('size').find('height').text)

    ymin, xmin, ymax, xmax = None, None, None, None

    ymin = int(boxes.find("bndbox/ymin").text)
    xmin = int(boxes.find("bndbox/xmin").text)
    ymax = int(boxes.find("bndbox/ymax").text)
    xmax = int(boxes.find("bndbox/xmax").text)

    xmin_corr = float(xmin / image_width)
    xmax_corr = float(xmax / image_width)
    ymin_corr = float(ymin / image_height)
    ymax_corr = float(ymax / image_height)

    unit_bbox = [xmin_corr, ymin_corr, xmax_corr, ymax_corr]
    valid_bbox = validate_bboxes(unit_bbox)
    all_bboxes.append(valid_bbox)

  assert len(labels) == len(all_bboxes)  # num. of labels must match num. of bboxes
  return img_name, labels, all_bboxes


def load_txt_annotations(csv_file: str) -> Tuple:
  """
  Load annotations and class labels from text files.
  Expected file structure: [class_name, x_min, y_min, x_max, y_max]
  x_min,y_min,x_max,y_max are bounding box coordinates already normalized
  relative to an image's respective width and height dimensions so the values
  expected should range between 0 and 1.
  """
  labels = []
  bboxes = []
  img_name = csv_file.split("/")[-1]  # get filename from path

  with open(csv_file, "r") as annot_file:
    lines = annot_file.readlines()
    for line in lines:
      data = line.split()
      labels.append(data[0])  # class label should be a string
      y_min = float(data[2])
      x_min = float(data[1])
      y_max = float(data[4])
      x_max = float(data[3])
      unit_bbox = [x_min, y_min, x_max, y_max]
      valid_bbox = validate_bboxes(unit_bbox)
      bboxes.append(valid_bbox)

  assert len(labels) == len(bboxes)  # num. of labels must match num. of bboxes
  return img_name, labels, bboxes


def create_image_df(data_dir, labels_dir, from_text_file=False):
  """
  Create image-label-annotations dataframe
  Args:
  	data_dir: image directory
  	labels_dir: labels/annotations directory
  	from_csv: indicates whether csv annotations or xml style annotations
  """
  data_dict = {"id": [], "image": [], "label": [], "bboxes": []}
  if from_text_file == False:

    img_annot_dict = {}
    image_paths = [data_dir + i for i in os.listdir(data_dir)]

    ## Create an image:annot mapping
    for i in image_paths:
      img_annot_dict[str(i)] = labels_dir + f"{i.split('/')[-1][:-4]}.xml"

    ## Extract labels and annotations from xml respective files
    for img_path, annot_path in img_annot_dict.items():
      image_name, labels, annotations = load_xml_annotations(annot_path)

      # create output image-label-annotation dataframe
      data_dict["id"].append(image_name[:-4])
      data_dict["image"].append(img_path)
      data_dict["label"].append(labels)
      data_dict["bboxes"].append(annotations)  # [xmin, ymin, xmax, ymax]

  else:
    img_annot_dict = {}
    image_paths = [data_dir + i for i in os.listdir(data_dir)]

    ## Create an image:annot mapping
    for i in image_paths:
      img_annot_dict[str(
          i
      )] = labels_dir + f"{i.split('/')[-1][:-4]}.txt"  # extract filename and link it to corresp. labels file

    ## Extract labels and annotations from xml respective files
    for img_path, annot_path in img_annot_dict.items():
      image_name, labels, annotations = load_txt_annotations(annot_path)

      # create output image-label-annotation dataframe
      data_dict["id"].append(image_name[:-4])
      data_dict["image"].append(img_path)
      data_dict["label"].append(labels)
      data_dict["bboxes"].append(annotations)  # [xmin, ymin, xmax, ymax]

  image_df = pd.DataFrame(data_dict)

  del data_dict
  gc.collect()

  return image_df


def create_segmentation_df(image_dir: str, masks_dir: str) -> pd.DataFrame:
  """
  Create an image, masked_image and labels dataframe for image
  segmentation data upload.
  Returns:
  	A dataframe with the id, image, label and mask columns.
  Customization Guide:
  	Update `data_dict` by appending the corresponding data to the values lists.
  	id: (str) used to uniquely identify each image.
  	image: (str) holds a list of image paths
  	label: (str) holds a list of labels,.i.e ['car', 'tv', 'car']
  	mask: (str) is a list of paths to images containing masks corresponding to each label
  	in the labels key of an image.i.e.
  		['car_masked_image_path', 'tv_masked_image_path', 'car_masked_image_path']
  	The order of the masks and labels lists matters and they must be of the same length.
  	Each masked image should therefore contain a single mask mapping to a single label.
  """
  data_dict = {"id": [], "image": [], "label": [], "mask": []}

  ## Your image processing code here

  segmentation_df = pd.DataFrame(data_dict)
  return segmentation_df



#! Clarifai data upload

import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import cpu_count

from base import Chunker
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from datasets import (ImageClassificationDataset, TextClassificationDataset,
                      VisualDetectionDataset, VisualSegmentationDataset)
from omegaconf import OmegaConf
from tqdm import tqdm

from clarifai_utils.client import create_stub


def upload_data(config, inputs, inp_stub):
  """
  Upload inputs to clarifai platform dataset.
  Args:
  	config: auth and data path info.
  	inputs: input protos
  	inp_stub: grpc stub
  """
  STUB = inp_stub
  USER_APP_ID = resources_pb2.UserAppIDSet(
      user_id=config.auth["user_id"], app_id=config.auth["app_id"])

  upload_count = 0
  retry_upload = []  # those that fail to upload are stored for retries

  for inp_proto in inputs:
    response = STUB.PostInputs(
        service_pb2.PostInputsRequest(user_app_id=USER_APP_ID, inputs=[inp_proto]),)

    if response.status.code != status_code_pb2.SUCCESS:
      try:
        print(f"Post inputs failed, status:\n{response.inputs[0].status.details}\n")
      except:
        print(f"Post inputs failed, status:\n{response.status.details}\n")
        retry_upload.append(inp_proto)
    else:
      upload_count += 1

  return retry_upload


def upload_annotations(config, inputs, inp_stub):
  """
  Upload image annotations to clarifai detection dataset
  """
  STUB = inp_stub
  USER_APP_ID = resources_pb2.UserAppIDSet(
      user_id=config.auth["user_id"], app_id=config.auth["app_id"])

  upload_count = 0
  retry_upload = []  # those that fail to upload are stored for retries

  for annot_proto in inputs:
    response = STUB.PostAnnotations(
        service_pb2.PostAnnotationsRequest(user_app_id=USER_APP_ID, annotations=[annot_proto]),)

    if response.status.code != status_code_pb2.SUCCESS:
      try:
        print(f"Post annotations failed, status:\n{response.annotations[0].status.details}\n")
      except:
        print(f"Post annotations failed, status:\n{response.status.details}\n")
        retry_upload.append(annot_proto)
    else:
      upload_count += 1

  return retry_upload


def concurrent_inp_upload(config, inputs, workers, chunks, stub):
  """
  Upload images concurrently for efficiency.
  """
  inp_threads = []
  retry_upload = []

  with ThreadPoolExecutor(max_workers=workers) as executor:
    for inp_batch in tqdm(inputs, total=chunks + 1, desc="uploading.."):
      inp_threads.append(executor.submit(upload_data, config, inp_batch, stub))
      time.sleep(0.2)

  for job in tqdm(
      as_completed(inp_threads), total=chunks + 1, desc="retry uploading failed protos..."):
    if job.result():
      retry_upload.extend(job.result())
  if len(list(retry_upload)) > 0:  ## TODO: use api_with_retries functionality via upload_data()
    _ = upload_data(config, retry_upload, stub)


def concurrent_annot_upload(config, inputs, workers, chunks, stub):
  """
  Upload annotations concurrently for efficiency.
  """
  annot_threads = []
  retry_annot_upload = []

  with ThreadPoolExecutor(max_workers=workers) as executor:
    for annot_batch in tqdm(inputs, total=chunks + 1, desc="uploading..."):
      annot_threads.append(executor.submit(upload_annotations, config, annot_batch, stub))
      time.sleep(0.2)

  for job in tqdm(
      as_completed(annot_threads), total=chunks + 1, desc="retry uploading failed protos..."):
    if job.result():
      retry_annot_upload.extend(job.result())
  if len(retry_annot_upload) > 0:
    ## TODO: use api_with_retries functionality via upload_annotations()
    _ = upload_annotations(config, retry_annot_upload, stub)


def upload_to_clarifai(config, task: str = "visual_clf"):
  """
  Execute data upload.
  Args:
    `config`: auth and data path info.
    `task`: Machine Learning domain task data type.
       Can be either of `visual_clf`, `visual_det` or `text_clf`.
  """
  STUB = create_stub()
  workers = cpu_count()

  if task == "text_clf":
    dataset_obj = TextClassificationDataset(config.data["clf_text_dir"], config.data["dataset_id"],
                                            config["split"])
    text_protos = dataset_obj._get_input_protos()
    text_protos = dataset_obj.to_list(text_protos)

    # Upload text
    chunks = len(text_protos) // workers
    chunked_text_protos = Chunker(text_protos, config["chunk_size"]).chunk()

    concurrent_inp_upload(config, chunked_text_protos, workers, chunks, STUB)

  elif task == "visual_det":
    dataset_obj = VisualDetectionDataset(
        config.data["visual_det_image_dir"],
        config.data["visual_det_labels_dir"],
        config.data["dataset_id"],
        config["split"],
        labels_from_text_file=False)
    img_protos, annotation_protos = dataset_obj._get_input_protos()
    img_protos = dataset_obj.to_list(img_protos)

    # Upload images
    chunks = len(img_protos) // workers
    chunked_img_protos = Chunker(img_protos, config["chunk_size"]).chunk()

    concurrent_inp_upload(config, chunked_img_protos, workers, chunks, STUB)

    # Upload annotations:
    print("Uploading annotations.......")
    annotation_protos = dataset_obj.to_list(annotation_protos)
    chunks_ = len(annotation_protos) // workers
    chunked_annot_protos = Chunker(annotation_protos, config["chunk_size"]).chunk()

    concurrent_annot_upload(config, chunked_annot_protos, workers, chunks_, STUB)

  elif task == "visual_seg":
    dataset_obj = VisualSegmentationDataset(config.data["visual_seg_image_dir"],
                                            config.data["visual_seg_masks_dir"],
                                            config.data["dataset_id"], config["split"])
    img_protos, mask_protos = dataset_obj._get_input_protos()
    img_protos = dataset_obj.to_list(img_protos)
    mask_protos = dataset_obj.to_list(mask_protos)

    # Upload images
    chunks = len(img_protos) // workers
    chunked_img_protos = Chunker(img_protos, config["chunk_size"]).chunk()

    concurrent_inp_upload(config, chunked_img_protos, workers, chunks, STUB)

    # Upload masks:
    print("Uploading masks.......")
    chunks_ = len(mask_protos) // workers
    chunked_mask_protos = Chunker(mask_protos, config["chunk_size"]).chunk()

    concurrent_annot_upload(config, chunked_mask_protos, workers, chunks_, STUB)

  else:
    dataset_obj = ImageClassificationDataset(config.data["clf_image_dir"],
                                             config.data["dataset_id"], config["split"])
    img_protos = dataset_obj._get_input_protos()
    img_protos = dataset_obj.to_list(img_protos)

    # Upload images
    chunks = len(img_protos) // workers
    chunked_img_protos = Chunker(img_protos, config["chunk_size"]).chunk()

    concurrent_inp_upload(config, chunked_img_protos, workers, chunks, STUB)


if __name__ == "__main__":
  yaml_path = "./config.yaml"
  config = OmegaConf.load(yaml_path)
  upload_to_clarifai(config, task=config["task"])



from typing import List

import pandas as pd
from base import ClarifaiDataset
from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct
from tqdm import tqdm
from utils import create_image_df, create_segmentation_df


class ImageClassificationDataset(ClarifaiDataset):
  """
  Load image classification data from csv file.
  The csv must have only two columns, the image path `image_path`
  and the `label` columns.
  """

  def __init__(self, csv_file_path: str, dataset_id: str, split: str) -> None:
    super().__init__(csv_file_path, dataset_id, split)
    self.image_df = pd.read_csv(self.csv_file_path)

  def __len__(self):
    return len(self._all_input_protos)

  def create_input_protos(self, image_path: str, label: str, input_id: str, dataset_id: str,
                          metadata: Struct):
    """
    Create input protos for each image, label input pair.
    Args:
    	`image_path`: full image path.
    	`label`: image label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: image metadata
    Returns:
    	An input proto representing a single row input
    """
    input_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            image=resources_pb2.Image(base64=open(image_path, 'rb').read(),),
            concepts=[
                resources_pb2.Concept(id=label, name=label, value=1.),
            ],
            metadata=metadata))

    return input_proto

  def _get_input_protos(self) -> List:
    """
    Creates input image protos for each row of the dataframe.
    Returns:
    	A list of input protos
    """
    for i, row in tqdm(self.image_df.iterrows(), desc="Loading images"):
      metadata = Struct()
      image_path = row["image_path"]
      label = row["label"]  # clarifai concept
      input_id = f"{self.split}-{str(i)}"
      metadata.update({"label": label, "split": self.split})

      input_proto = self.create_input_protos(image_path, label, input_id, self.dataset_id,
                                             metadata)

      self._all_input_protos.append(input_proto)

    return iter(self._all_input_protos)


class TextClassificationDataset(ClarifaiDataset):
  """
  Upload text classification datasets to clarifai datasets
  """

  def __init__(self, csv_file_path: str, dataset_id: str, split: str) -> None:
    super().__init__(csv_file_path, dataset_id, split)
    self.image_df = pd.read_csv(self.csv_file_path)

  def __len__(self):
    return len(self._all_input_protos)

  def create_input_protos(self, text_input: str, label: str, input_id: str, dataset_id: str,
                          metadata: Struct):
    """
    Create input protos for each text, label input pair.
    Args:
    	`text_input`: text string.
    	`label`: text label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: text metadata
    Returns:
    	An input proto representing a single row input
    """
    input_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            text=resources_pb2.Text(raw=text_input),
            concepts=[resources_pb2.Concept(id=label, name=label, value=1.)],
            metadata=metadata))

    return input_proto

  def _get_input_protos(self) -> List:
    """
    Creates input protos for each row of the dataframe.
    Returns:
    	A list of input protos
    """
    for i, row in tqdm(self.image_df.iterrows(), desc="Loading Text"):
      metadata = Struct()
      text = row["text"]
      label = row["label"]  # clarifai concept
      input_id = f"{self.split}-{str(i)}"
      metadata.update({"label": row["label"], "split": self.split})

      input_proto = self.create_input_protos(text, label, input_id, self.dataset_id, metadata)

      self._all_input_protos.append(input_proto)

    return iter(self._all_input_protos)


class VisualDetectionDataset(ClarifaiDataset):
  """
  Visual detection dataset upload class.
  """

  def __init__(self, file_path, label_path, dataset_id, split,
               labels_from_text_file=False) -> None:
    super().__init__(None, dataset_id, split)
    self.file_path = file_path
    self.label_path = label_path
    self.image_df = create_image_df(
        self.file_path, self.label_path, from_text_file=labels_from_text_file)
    self._annotations = []

  def create_input_protos(self, image_path: str, label: List, input_id: str, dataset_id: str,
                          metadata: Struct):
    """
    Create input protos for each image, label input pair.
    Args:
    	`image_path`: absolute file path to image
    	`label`: image label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: image metadata
    Returns:
    	An input proto representing a single row input
    """
    input_image_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            image=resources_pb2.Image(base64=open(image_path, 'rb').read(),),
            concepts=[
                resources_pb2.Concept(id=_label, name=_label, value=1.) for _label in set(label)
            ],
            metadata=metadata))

    return input_image_proto

  def create_annotation_proto(self, label: str, annotations: List, input_id: str, dataset_id: str):
    """
    Create an input proto for each bounding box, label input pair.
    Args:
    	`label`: annotation label
    	`annotations`: a list of a single bbox's coordinates.
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    Returns:
    	An input proto representing a single row input
    """
    input_annot_proto = resources_pb2.Annotation(
        input_id=input_id,
        data=resources_pb2.Data(regions=[
            resources_pb2.Region(
                region_info=resources_pb2.RegionInfo(bounding_box=resources_pb2.BoundingBox(
                    # Annotations ordering: [xmin, ymin, xmax, ymax]
                    # top_row must be less than bottom row
                    # left_col must be less than right col
                    top_row=annotations[1],  #y_min
                    left_col=annotations[0],  #x_min
                    bottom_row=annotations[3],  #y_max
                    right_col=annotations[2]  #x_max
                )),
                data=resources_pb2.Data(concepts=[resources_pb2.Concept(id=label, value=1.)]))
        ]))

    return input_annot_proto

  def _get_input_protos(self) -> List:
    """
    Create input image protos for each row of the dataframe.
    Returns:
    	A list of input protos
    """
    for i, row in tqdm(self.image_df.iterrows(), desc="Loading images"):
      metadata = Struct()
      image = row["image"]
      label = row["label"]  # list
      bboxes = row["bboxes"]  # [xmin, ymin, xmax, ymax]
      input_id = f"{row['id']}-{str(i)}"
      metadata.update({"label": row['id'], "split": self.split})

      input_image_proto = self.create_input_protos(image, label, input_id, self.dataset_id,
                                                   metadata)
      self._all_input_protos.append(input_image_proto)

      # iter over bboxes and annotations and map to each input_id
      # one id could have more than one bbox and label
      for i in range(len(bboxes)):
        input_annot_proto = self.create_annotation_proto(label[i], bboxes[i], input_id,
                                                         self.dataset_id)
        self._annotations.append(input_annot_proto)

    return iter(self._all_input_protos), iter(self._annotations)


class VisualSegmentationDataset(ClarifaiDataset):
  """
  Visual segmentation dataset upload class.
  """

  def __init__(self, file_path, masks_path, dataset_id, split) -> None:
    super().__init__(None, dataset_id, split)
    self.file_path = file_path
    self.masks_path = masks_path
    self.image_df = create_segmentation_df(self.file_path, self.masks_path)
    self._masks = []

  def create_input_protos(self, image_path: str, label: List, input_id: str, dataset_id: str,
                          metadata: Struct):
    """
    Create input protos for each image, label input pair.
    Args:
    	`image_path`: absolute image file path
    	`label`: image label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: image metadata
    Returns:
    	An input proto representing a single row input
    """
    input_image_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            image=resources_pb2.Image(base64=open(image_path, 'rb').read(),),
            concepts=[
                resources_pb2.Concept(id=_label, name=_label, value=1.) for _label in set(label)
            ],
            metadata=metadata))

    return input_image_proto

  def create_mask_proto(self, label: str, mask: str, input_id: str, dataset_id: str):
    """
    Create an input mask proto for an input masked image and label.
    Args:
    	`label`: image label
    	`mask`: masked image path
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    Returns:
    	An input proto representing a single row input
    """
    input_mask_proto = resources_pb2.Annotation(
        input_id=input_id,
        data=resources_pb2.Data(regions=[
            resources_pb2.Region(
                region_info=resources_pb2.RegionInfo(mask=resources_pb2.Mask(
                    image=resources_pb2.Image(base64=open(mask, 'rb').read()))),
                data=resources_pb2.Data(concepts=[resources_pb2.Concept(id=label, value=1.)]))
        ]))

    return input_mask_proto

  def _get_input_protos(self) -> List:
    """
    Create input image protos for each row of the dataframe.
    Returns:
    	A list of input protos
    """
    for i, row in tqdm(self.image_df.iterrows(), desc="Loading images"):
      metadata = Struct()
      image = row["image"]  # image path
      label = row["label"]  # list of class labels
      masks = row["mask"]  # list of image paths with masked regions
      input_id = f"{row['id']}-{str(i)}"
      metadata.update({"label": row['id'], "split": self.split})

      input_image_proto = self.create_input_protos(image, label, input_id, self.dataset_id,
                                                   metadata)
      self._all_input_protos.append(input_image_proto)

      ## Iterate over each masked image and create a proto for upload to clarifai
      ## The length of masks and labels must be equal
      for i, mask in enumerate(masks):
        input_mask_proto = self.create_mask_proto(label[i], mask, input_id, self.dataset_id)
        self._masks.append(input_mask_proto)

    return iter(self._all_input_protos), iter(self._masks)
